/* procttl.c
**
** set the ps-title of a sub-process.
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/param.h>
#include <limits.h>

#include "mrmacs.h"

/* SPT_ARGVSTR (setproctitle by changing the buffer containing the arguments
**              of the running program) is the default, so it should be
**              defined here.
*/
#define SPT_ARGVSTR

/* if a system call `pstat' exists, we must include the corresponding
** header file and remove the previous definition of SPT_ARGVSTR
*/
#ifdef SPT_PSTAT
#include <sys/pstat.h>
#undef SPT_ARGVSTR
#endif

/* on a MIPS-alike system, there is a system call for setting the process
** title. We must include the corresponding header file(s) and remove the
** definition of `SPT_ARGVSTR'.
*/
#ifdef SPT_SYSMIPS
#include <sys/sysmips.h>
#include <sys/sysnews.h>
#undef SPT_ARGVSTR
#endif

/* on an SCO-alike system (and maybe some other - older - UNIX systems) the
** only way for setting the process title is by changing the contents of
** the corresponding kernel memory entry. In this case, we must include
** some header files and remove the definition of `SPT_ARGVSTR'.
*/
#ifdef SPT_SCO
#include <sys/immu.h>
#include <sys/dir.h>
#include <sys/user.h>
#include <sys/fs/s5param.h>
#if PSARGSZ > MAXLINE
#define SPT_BUFSIZE PSARGSZ
#endif
#undef SPT_ARGVSTR
#endif

/* on some BSD-alike systems ther seems to exists a global variable (or such)
** which holds a pointer to the argument vector and the argument counter.
** if the corresponding system doesn't define the macro PS_STRINGS,
** then fall back to the default by removing `SPT_PSSTRINGS'; otherwise remove
** the definition of `SPT_ARGVSTR'.
*/
#ifdef SPT_PSSTRINGS
#include <machine/vmparam.h>
#include <sys/exec.h>
#ifdef PS_STRINGS
#ifndef NKPDE
#define NKPDE 63
typedef unsigned int *pt_entry_t;
#endif
#undef SPT_ARGVSTR
#else
#undef SPT_PSSTRINGS
#endif
#endif

/* on some systems, changing the normal argument vector seems to be sufficient
** for changing the process title. Remove `SPT_ARGVSTR' in this case, too.
*/
#ifdef SPT_ARGV
#undef SPT_ARGVSTR
#endif

/* last not least, some systems seem to have a library function or sytem call
** for changing the process title itself. Remove `SPT_ARGVSTR' in this case,
** too.
*/
#ifdef SPT_BUILTIN
#undef SPT_ARGVSTR
#endif

/* define `SPT_BUFSIZE' if it's not already defined */
#ifndef SPT_BUFSIZE
#define SPT_BUFSIZE MAXLINE
#endif

#include "tools.h"

static char *oldargs = NULL, **oldargv;
static char *endargs = NULL;

/* Diese Funktion wird fr die Vorarbeit bei solchen System (wie z.B. Linux)
** bentigt, die als einzige Mglichkeit zur nderung der `ps'-Ausgabe das
** berschreiben des Zeichenpuffers, der die Argumente und Umgebungsvariablen
** enthlt, vorsehen.
*/
void init_procttl (int argc, char **argv, char **envp)
{
#if defined(SPT_ARGV) || defined(SPT_ARGVSTR)
    int i;
    size_t envsize = 0, envlast = 0;
    extern char **environ;
#if 0
    char *envbuf, *p;
#endif

    /* Anscheinend werden die `argv'-Zeichenketten und die fr die
    ** Umgebungsvariablen direkt hintereinander abgelegt. Wenn ich die
    ** Umgebungsvariablen an eine neue Stelle verschiebe, dann kann ich
    ** - nach dem Einlesen der Argumente - diesen Platz und den Platz
    ** fr die `argv'-Zeichenketten als Platz fr meine eigenen Texte nutzen.
    ** Dies scheint brigens die einzige Mglichkeit zu sein, unter Linux
    ** die Anzeige eines `ps' fr das jeweilige Programm zu ndern.
    */

    /* Also: die Umgebungsvariablen werden zunchst in einen neuen
    ** Speicherbereich verschoben.
    */

#ifndef SPT_ARGV
    for (i = 0; envp[i]; ++i) { ++envlast; envsize += strlen (envp[i]) + 1; }
    environ = (char **) galloc (__FILE__, __LINE__, (envlast+1)*sizeof(char *));
#if 0
    envbuf = galloc (__FILE__, __LINE__, envsize + 10);
    p = envbuf;
#endif
    for (i = 0; noNULL(envp[i]); ++i) {
#if 0
	environ[i] = p;
	strcpy (p, envp[i]);
	p += strlen (p) + 1;
#else
	environ[i] = gdupl (__FILE__, __LINE__, envp[i]);
#endif
    }
#if 0
    *p++ = '\0'; *p++ = '\0'; *p++ = '\0'; *p++ = '\0';
#endif
    environ[envlast] = NULL;
#endif

    /* So, die Umgebungsvariablen sind also in einen neuen (zusammenhngenden)
     * Speicherbereich gewandert.
     * Jetzt mu festgestellt werden, wieviel Platz die `argv'-Zeichenketten
     * und die Zeichenketen aus dem ursprnglichen Environment zusammengenommen
     * bieten.
     *
     * Zuerst mal den Argumentvektor:
     */
    oldargv = argv;
#ifndef SPT_ARGV
    oldargs = *argv;
    endargs = oldargs + strlen (oldargs);
    for (i = 1; i < argc; ++i)
    {
	if (endargs + 1 == argv[i]) { endargs = argv[i] + strlen (argv[i]); }
    }

    /* Jetzt die Umgebungsvariablen: */

    for (i = 0; envp[i]; ++i) {
	if (endargs + 1 == envp[i]) { endargs = envp[i] + strlen (envp[i]); }
    }
#endif
#else
    ;
#endif
    /* So, das wr's erst einmal! */
}

#ifndef SPT_BUILTIN
/* `setproctitle' ist im wesentlichen dem `sendmail'-Quellcode entnommen;
** allerdings habe ich ein paar kleine nderungen vorgenommen, was die
** Konfiguration betrifft. Dieser Teil wird nur dann bersetzt, wenn
** `setproctitle' nicht bereits als Bibliotheksfunktion oder Systemcall zur
** verfgung steht.
*/
void setproctitle (char *fmt, ...)
{
    va_list ap;
#ifdef SPT_PSTAT
    union pstun pst;
#endif
#ifndef SPT_ARGVSTR
    static char cmdttl[SPT_BUFSIZE];
#endif
#ifdef SPT_SCO
    off_t seek_off;
    static int kmem = -1;
    static int kmempid = -1;
    struct user u;
#endif
#ifndef SPT_ARGVSTR
    memset (cmdttl, '\0', SPT_BUFSIZE);
    va_start (ap, fmt);
    vsnprintf (cmdttl, SPT_BUFSIZE, fmt, ap);
    va_end (ap);
#endif
#ifdef SPT_PSTAT
    /* Anscheinend gibt es einige Systeme, die ber so etwas wie einen
    ** `pstat'-Systemcall verfgen. Mit diesem lt sich die `ps'-Ausgabe
    ** steuern.
    */
    pst.pst_command = cmdttl;
    pstat (PSTAT_SETCMD, pst, strlen (cmdttl), 0, 0);
#elif defined(SPT_SYSMIPS)
    /* Die MIPS-kompatiblen Systeme haben ebenfalls einen Systemcall fr
    ** so eine Aufgabe.
    */
    sysmips(SONY_SYSNEWS, NEWS_SETPSARGS, cmdttl);
#elif defined(SPT_SCO)
    /* Bei SCO-hnlichen Systemen ist eine nderung der `ps'-Ausgabe
    ** anscheinend nur ber eine direkte nderung des `kernel memory' mglich.
    ** Damit das richtig funktioniert, mu der betreffende Proze `root'-Rechte
    ** haben.
    */
    if (kmem < 0 || kmempid != getpid()) {
	if (kmem >= 0) {  close (kmem); }
	kmem = open ("/dev/kmem", O_RDWR, 0);
	if (kmem < 0) { return; }
	fcntl (kmem, F_SETFD, 1);
	kmempid = getpid();
    }
    buf[127] = '\0';
    seek_off = UVUBLK + (off_t) u.u_psargs - (off_t) &u;
    if (lseek (kmem, seek_off, SEEK_SET) == seek_off) {
	write (kmem, buf, SPT_BUFSIZE);
    }
#elif defined(SPT_PSSTRINGS)
    /* Einige BSD-artige Systeme verfgen anscheinend ber eine globale
    ** Datenstruktur, in der die fr eine `ps'-Ausgabe relevanten initialen
    ** `argc'- bzw. `argv'-Komponenten stehen.
    */
    PS_STRINGS->ps_nargvstr = 1;
    PS_STRINGS->ps_argvstr = cmdttl;
#elif defined(SPT_ARGV)
    /* Ich wei zwar nicht, welche Systeme dies sind; aber anscheinend mssen
    ** bei einigen nur die Zeiger im initialen `argv'-Vektor umgestellt werden,
    ** um eine nderung der `ps'-Ausgabe zu erreichen.
    */
    oldargv[0] = cmdttl;
    oldargv[1] = NULL;
#elif defined(SPT_ARGVSTR)
    /* Einige Systeme - darunter auch Linux - bieten nur die Mglichkeit, den
    ** Zeichenpuffer, auf den smtliche `argv'- und `environ'-Komponenten
    ** zeigen, zu verndern, um die `ps'-Ausgabe zu verndern. Dies wre nicht
    ** weiter problematisch, wre die Gre dieses Puffers bekannt; so mu sie
    ** stattdessen in einer eigenen Routine, die - noch vor der Auswertung
    ** irgendwelcher Kommandozeilenargumente - aufgerufen wird, bestimmt
    ** werden; anschlieend mssen die Argumente und Umgebungsvariablen in
    ** einen neuen Speicherbereich kopiert und die `argv'- und `environ'-
    ** Komponenten auf diesen neuen Puffer umgesetzt werden. Danach steht der
    ** ursprngliche Puffer fr die nderung der `ps'-Ausgabe zur Verfgung.
    */
    va_start (ap, fmt);
    memset (oldargs, '\0', (endargs - oldargs));
    vsnprintf (oldargs, (endargs - oldargs) + 1, fmt, ap);
    va_end (ap);
    oldargv[0] = oldargs;
    oldargv[1] = NULL;
#endif
}
#endif
